/*  $Id: linkd.c,v 1.5 2000/01/12 03:44:44 shaggy Exp $
 *      directory link utility
 *      
 *      relative path: ./fconfig
 *  
 *  Copyright (c) 1995-1998 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/members/shaggy
 */

#define __NO_STRING_INLINES	/* avoid some gnu clashes */

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>

typedef char sz;
typedef const char csz;
typedef int bool;
enum { true=1, false=0 };
#define nelem(a)	(sizeof(a)/sizeof(*(a)))
#define nil ((void*)0)

/* Note: this program needs some major rewriting, reorganizing, and
 * extending so it doesn't have static limits-- right now it's a quick
 * and dirty hack to get the job done (converted very quickly from
 * a shell script to a C program, without much restructuring)
 */


/* it seems that some systems (like old Ultrix) don't have these defined,
 * even though they're in the libraries
 * (hopefully we don't clash...)
 */
sz* strdup(csz*);
int symlink(csz*, csz*);


#ifndef PATH_MAX
#define PATH_MAX	1024	/* pathname max size */
#endif
#ifndef NAME_MAX
#define NAME_MAX	256	/* filename max size */
#endif


static csz *badpath;

static csz*
fullpathname(csz *d)
{
	sz oldwd[PATH_MAX];
	static sz wd[PATH_MAX];

	if(!getcwd(oldwd, sizeof oldwd)) {
		badpath = ".";
		return nil;
	} 
	if(chdir(d) < 0) {
		badpath = d;
		return nil;
	}
	if(!getcwd(wd, sizeof wd)) {
		badpath = d;
		return nil; 
	}
	if(chdir(oldwd) < 0) {
		badpath = ".";
		return nil;
	}
	return wd;
}

static csz*
relpath(csz *pdest, csz *psrc)
{
	sz src[PATH_MAX];
	sz dest[PATH_MAX];
	static sz relprefix[PATH_MAX];
	sz relpath[PATH_MAX];
	char *r;

	if(!(psrc = fullpathname(psrc)))
		return nil;
	sprintf(src, "%s/", psrc);

	if(!(pdest = fullpathname(pdest)))
		return nil;
	sprintf(dest, "%s/", pdest);

	relprefix[0] = '\0';
	relpath[0] = '\0';
	while(!strstr(dest, src)) {
		strcat(relprefix, "../");
		src[strlen(src)-1] = '\0';
		strrchr(src, '/')[1] = '\0';
	}
	if(strcmp(src, "/") == 0)
		strcpy(relprefix, "/");
	while(strcmp(src, dest) != 0) {
		char tmp[PATH_MAX];
		dest[strlen(dest)-1] = '\0';
		r = strrchr(dest, '/');
		strcpy(tmp, relpath);
		sprintf(relpath, "%s/%s", r+1, tmp);
		r[1] = '\0';
	}
	strcat(relprefix, relpath);
	while((r = strstr(relprefix, "//")))
		strcpy(r, r+1);
	if(strlen(relprefix) > 1)
		relprefix[strlen(relprefix)-1] = '\0';
	return relprefix;
}

int
main(int argc, csz *argv[])
{
	csz *argv0 = argv[0];
	csz *srcdir[32];
	sz sds[32][PATH_MAX];
	sz sd[PATH_MAX];
	csz *dest;
	int nsrcdir;
	int nsds;
	bool recurse;
	bool update;
	bool verbose;
	DIR *dir;
	struct dirent *de;
	int i;
	sz switches[20];

	if(strlen(argv0) >= sizeof("relpath")-1	
			&& strcmp(argv0+strlen(argv0)
			-(sizeof("relpath")-1), "relpath") == 0) {
		if(argc != 3) {
			fprintf(stderr, "usage: relpath <to> <from>\n");
			return 1;
		}
		dest = relpath(argv[1], argv[2]);
		if(dest) {
			printf("%s\n", dest);
			return 0;
		} else {
			perror(badpath);
			return 1;
		}
	}
	recurse = false;
	update = false;
	verbose = false;
	nsrcdir = 0;
	dest = nil;
	switches[0] = '\0';

	while(argc-- > 1) {
		++argv;
		if((*argv)[0] == '-') {
			sprintf(switches+strlen(switches), "%s ", *argv);
			switch((*argv)[1]) {
			case 'r':
				recurse = true;
				break;
			case 'u':
				update = true;
				break;
			case 'v':
				verbose = true;
				break;
			default:
				fprintf(stderr, "invalid switch: %s\n", *argv);
				exit(1);
			}
		}
		else {
			if(dest)
				srcdir[nsrcdir++] = dest;
			dest = *argv;
		}
	}
	if(!dest) {
		fprintf(stderr, "usage: linkd [-ruv] [<src>...] <dest>\n");
		return 1;
	}

	for(i=0; i<nelem(sds); i++)
		sds[i][0] = '\0';
	nsds = 0;

	if(!(dir = opendir(dest))) {
		perror(dest);
		return 1;
	} 
	while((de = readdir(dir))) {
		csz *s = de->d_name;
		if(s[0] == '.' && s[1] == 's' && isdigit(s[2])) {
			if(update)
				strcpy(sds[atoi(&s[2])], s);
			else
				unlink(s);
		}
	}
	closedir(dir);
	for(i=0; i<nelem(sds); i++)
		if(sds[i][0]) {
			if(i > nsds)
				strcpy(sds[nsds], sds[i]);
			++nsds;
		}

	if(verbose) {
		sz vsrcdir[2*nelem(sds)*PATH_MAX];
		sz vdest[2*PATH_MAX];
		csz *fp;
		vsrcdir[0] = '\0';
		for(i=0; i<nsrcdir; i++) {
			if(!(fp = fullpathname(srcdir[i]))) {
				perror(badpath);
				exit(1);
			}
			sprintf(vsrcdir+strlen(vsrcdir), "%s (%s) ", srcdir[i], fp);
		}
		if(!(fp = fullpathname(dest))) {
			perror(badpath);
			exit(1);
		}
		sprintf(vdest, "%s (%s)", dest, fp);
		if(update) {
			if(nsrcdir == 0) {
				if(nsds > 0)
					printf("Updating %s\n", vdest);
			} else
				printf("Updating and linking %s to %s\n",
					vsrcdir, vdest);
		} else
			printf("Linking %s to %s\n", vsrcdir, vdest);
	}

	for(i=0; i<nsrcdir; i++) {
		csz *rp;
		int n = 0;
		if(nsds > 0)
			n = atoi(&sds[nsds-1][2])+1;
		sprintf(sds[nsds], ".s%d", n);
		sprintf(sd, "%s/%s", dest, sds[nsds]);
		++nsds;
		unlink(sd);
		if(!(rp = relpath(srcdir[i], dest))) {
			perror(badpath);
			exit(1);
		}
		symlink(rp, sd);
	}

	for(i=0; i<nsds; i++) {
		sprintf(sd, "%s/%s", dest, sds[i]);
		if(!(dir = opendir(sd))) {
			perror(sd);
			return 1;
		}
		while((de = readdir(dir))) {
			sz f[PATH_MAX];
			sz file[PATH_MAX];
			sz destf[PATH_MAX];
			struct stat st;

			strcpy(f, de->d_name);
			if(strcmp(f, ".") == 0 || strcmp(f, "..") == 0)
				continue;
			if(f[0] == '.' && f[1] == 's' && isdigit(f[2]))
				continue;
			if(strcmp(f, "CVS") == 0)
				continue;
			if(strcmp(f, "SCCS") == 0)
				continue;
			if(strcmp(f, "RCS") == 0)
				continue;
			sprintf(file, "%s/%s", sd, f);
			sprintf(destf, "%s/%s", dest, f);
			if(stat(file, &st) < 0) 
				continue;
			if(S_ISDIR(st.st_mode)) {
				bool newdir;
				/*
				DIR *dir2;
				bool linked = false;
				if(!(dir2 = opendir(file))) {
					perror(file);
					return 1;
				}
				while((de = readdir(dir2))) {
					csz *s = de->d_name;
					if(s[0] == '.' && s[1] == 's'
							&& isdigit(s[2])) {
						linked = true;
						break;
					}
				}
				closedir(dir2);
				if(linked)
					continue;
				*/
	
				if(stat(destf, &st) < 0)
					st.st_mode = 0;
				if(S_ISREG(st.st_mode)) {
					fprintf(stderr, "linkd: refusing to clobber %s with directory\n", destf);
					return 1;
				}
				if((newdir = !S_ISDIR(st.st_mode))) {
					int n;
					mkdir(destf, 0777);
					for(n=0; n<nsds; n++) {
						sz dsf[PATH_MAX];
						sz dfs[PATH_MAX];
						csz *rp;
						sprintf(dsf, "%s/%s/%s", dest, sds[n], f);
						sprintf(dfs, "%s/%s/%s", dest, f, sds[n]);
						if((rp = relpath(dsf, destf)))
							symlink(rp, dfs);
					}
				}
				if(recurse) {
					int n;
					sz s[(2+nelem(sds))*PATH_MAX];
					sprintf(s, "%s %s", argv0, switches);
					if(!newdir)
						for(n=0; n<nsds; n++) 
							sprintf(s+strlen(s), " %s/%s/%s",
								dest, sds[n], f);
					sprintf(s+strlen(s), " %s", destf);
					if(system(s) < 0) {
						perror(s);
						return 1;
					}
				}
			} else if(S_ISREG(st.st_mode)) {
				sz df[PATH_MAX];
				sprintf(df, "%s/%s", sds[i], f);
				if(stat(destf, &st) < 0)
					st.st_mode = 0;
				if(S_ISLNK(st.st_mode)) {
					unlink(destf);
					symlink(df, destf);
				} else if(!S_ISREG(st.st_mode)) 
					symlink(df, destf);
			}
		}
		closedir(dir);
	}
	return 0;
}

