/*  $Id: Elf32.cc,v 1.11 2000/01/12 03:44:44 shaggy Exp $
 *      relative path: lib/bff/elf32
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/members/shaggy
 */

#pragma Make Includes_Banned += -nostdinc -nostdinc++ 
#pragma Make Flags_Banned += -static 

#define USE_STDLIB
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <vstream>
#include <vdir>
#include <vthread>
#include <Error>
#include "Elf32.hh"


void
Elf32::read(Phdr &phdr, int i) const
{
	stream->read_at(ehdr.e_phoff+i*ehdr.e_phentsize, ADR_LEN(phdr));
}

void
Elf32::read(Shdr &shdr, int i) const
{
	stream->read_at(ehdr.e_shoff+i*ehdr.e_shentsize, ADR_LEN(shdr));
}


Elf32::Elf32(vstream *newstream) 
{
	stream = newstream;
	stream->read_at(0, ADR_LEN(ehdr));
	if(    ehdr.e_ident[EI_MAG0] != MAG0
	    || ehdr.e_ident[EI_MAG1] != MAG1
	    || ehdr.e_ident[EI_MAG2] != MAG2
	    || ehdr.e_ident[EI_MAG3] != MAG3)
		THROW(Error("Not an ELF file"));
	link = 0;
	shstrtab = 0;
}

Elf32::~Elf32()
{
	delete stream;
}


ulong
Elf32::hash(const uchar *name) 
{
	ulong h = 0, g;
	while(*name) {
		h = (h << 4) + *name++;
		if ((g = h & 0xf0000000))
			h ^= g >> 24;
		h &= ~g;
	}
	return h;
}


Elf32::Word
Elf32::link_info::symhash(const char *name, Word x)
{
	Word nbucket = hash[0];
	Word *bucket = &hash[2];
	Word *chain = &hash[2+nbucket];
	Word y = bucket[x%nbucket];

	while(y != STN_UNDEF) {
		if(!strcmp(&strtab[symtab[y].st_name], name))
			return y;
		y = chain[y];
	}
	return y;
}

Elf32::Sym*
Elf32::dyn_chain::lookup_level(int level,
		const char *name, Word hnum, link_info **linkp)
{
	Sym* sym = 0;
	dyn_chain *chain = this;
	bool hit_bottom = true;

	printf("level=%d\n", level);
	while(chain && !sym) {
		if(level > 0) {
			if(chain->next_level) {
				sym = chain->next_level->lookup_level(
					level-1, name, hnum, linkp);
				hit_bottom = false;
			}
		} else {
			link_info *link = chain->link;
			if(!link->initialized && !link->initializing)	
				link->initialize(); 
			Word snum = link->symhash(name, hnum);
			printf("snum=%ld\n", snum);
			if(snum != STN_UNDEF) {
				// link->elf32->dump(link->symtab[snum], link->strtab);
				if(link->symtab[snum].st_shndx != SHN_UNDEF) {
					sym = &link->symtab[snum];
					*linkp = link;
				}	
			}
		}
		chain = chain->next; 
	}
	if(level && hit_bottom)
		return (Sym*)-1;
	return sym;
}


Elf32::Word
Elf32::dyn_chain::lookup(const char *name)
{
	// XXX -- need way of dealing with weak symbols
	Word hnum = Elf32::hash(name);
	Sym *sym;
	int level = 0;
	link_info *link;

	while(!(sym = lookup_level(level, name, hnum, &link)))
		level++;
	printf("got sym=%p\n", sym);
	if(sym == (Sym*)-1) {
		// XXX -- these should be probably be set to something useful
//		if(!strcmp(name, "___brk_addr"))
//			return (Word)sbrk(0);
//		if(!strcmp(name, "__environ"))
//			return 0;
		printf("sym not found. chain=\n");
//		dump();
		THROW(Error(name));
		return 0;
	}
	if(sym->st_shndx == SHN_ABS)  
		return sym->st_value;
	else
		return sym->st_value + link->loadaddr;
}


Elf32::Word
Elf32::link_info::symval(Word snum)
{
	Sym &s = symtab[snum];

	printf("sym=%lx\n", snum);
	// elf32->dump(s, strtab);
	if(s.st_shndx == SHN_UNDEF) {
		// XXX -- can this be optimized to save the
		//	  value after it's looked up,
		//	  or will that be a problem when there
		//	  are weak bindings, etc... ?
		const char *name = &strtab[s.st_name];
		printf("look up %s\n", name);
		return chain_top->lookup(name);
	}
	if(s.st_shndx == SHN_ABS) 
		return s.st_value;
	else
		return s.st_value + loadaddr;
}



// XXX/DEP -- I386 (feb)
Elf32::Addr
Elf32::link_info::reloc(const Rel &r, Sword addend, bool bind_jmp)
{
	Word *ofs = (Word*)(r.r_offset + addend);
	switch(r.r_type()) {
	case R_386_NONE:
		break;
	case R_386_GLOB_DAT:	
		*ofs = symval(r.r_sym());
		break;
	case R_386_JMP_SLOT:
		if(bind_jmp)
			*ofs = symval(r.r_sym());
		else
			*ofs += addend;
		break;
	case R_386_32:
		*ofs += symval(r.r_sym());
		break;
	case R_386_PC32:	
		*ofs += symval(r.r_sym()) - (Addr)ofs;
		break;
	case R_386_RELATIVE:
		*ofs += addend;
		break;
	case R_386_COPY:
	case R_386_GOTOFF:
	case R_386_GOTPC:
	case R_386_GOT32:
	case R_386_PLT32:	
	default:
		printf("Warning: %s not supported\n",
				r_type_str(r.r_type()));
		break;
	}
	printf("*(%p) = %lx\n", ofs, *ofs);
	return (Addr)ofs;
}


Elf32::Addr
Elf32::link_info::resolver(Word ofs)
{
	printf("resolving (%x) ofs:%ld\n", (uint)this, ofs);
	if(pltrel == DT_REL)
		return reloc(*(Rel*)(jmprel+ofs), loadaddr);
	else if(pltrel == DT_RELA)
		return reloc(*(Rel*)(jmprel+ofs),
			((Rela*)(jmprel+ofs))->r_addend);
	THROW(Error("unexpected symbol type"));
}


// XXX/DEP -- I386 (feb)
#ifdef ARCH_i386
#define MANGLED_resolve	"resolve__Q25Elf329link_info"
#define MANGLED_resolver "resolver__Q25Elf329link_infoUl"
__asm__(".globl "MANGLED_resolve";" 
	MANGLED_resolve ": "
#ifdef __PIC__
		"call "MANGLED_resolver"@PLT;"
#else
		"call "MANGLED_resolver";"
#endif
          	"addl $8,%esp; "
         	"jmp *(%eax)"
);
#endif


Elf32::link_info::link_info(Word adjust, const Dyn *dynamic,
					bool preinitialized)
{
	memset(this, 0, sizeof(*this));
	loadaddr = adjust;
	dynamic_addr = dynamic;
	for(const Dyn *dp = dynamic; dp->d_tag != DT_NULL; dp++) {
		const Dyn &d = *dp;
		switch(d.d_tag) {
		case DT_PLTRELSZ:
			pltrelsz = d.d_val; break;
		case DT_PLTGOT:
			pltgot = (Addr*)(d.d_ptr + adjust); break;
		case DT_HASH:
			hash = (Word*)(d.d_ptr + adjust); break;
		case DT_STRTAB:
			strtab = (const char*)(d.d_ptr + adjust); break;
		case DT_SYMTAB:
			symtab = (Sym*)(d.d_ptr + adjust); break;
		case DT_RELA:
			rela = d.d_ptr + adjust; break;
		case DT_RELASZ:	  relasz = d.d_val; break;
		case DT_RELAENT:  relaent = d.d_val; break;
		case DT_STRSZ:	  strsz = d.d_val; break;
		case DT_SYMENT:	  syment = d.d_val; break;
		case DT_INIT:
			init = (void (*)())(d.d_ptr + adjust); break;
		case DT_FINI:
			fini = (void (*)())(d.d_ptr + adjust); break;
		case DT_SYMBOLIC: symbolic = true; break;
		case DT_REL:
			rel = d.d_ptr + adjust; break;
		case DT_RELSZ:	  relsz = d.d_val; break;
		case DT_RELENT:	  relent = d.d_val; break;
		case DT_PLTREL:	  pltrel = d.d_val; break;
		case DT_TEXTREL:  textrel = true; break;
		case DT_JMPREL:
			jmprel = d.d_ptr + adjust; break;
		case DT_DEBUG:
			debug = (Debug*)(d.d_ptr + adjust); break;
		}
	}
	if(!strtab)
		THROW(Error("No dynamic string table"));
	for(const Dyn *dp = dynamic; dp->d_tag != DT_NULL; dp++) {
		const Dyn &d = *dp;
		switch(d.d_tag) {
		case DT_SONAME:
			libname = &strtab[d.d_val];
			break;
		}
	}
	initialized = preinitialized;
}

void
Elf32::link_info::load_dependencies()
{
	bool libc_needed = true;

	printf("** loading depend for %s\n", libname);
	printf("chain=%p\n", chain);
	for(const Dyn *dp = dynamic_addr; dp->d_tag != DT_NULL; dp++) {
		const Dyn &d = *dp;
		switch(d.d_tag) {
		case DT_RPATH:
			printf("Warning: RPATH not supported (%s)\n",
				&strtab[d.d_val]);
			break;
		case DT_NEEDED:
			printf("we(%s) need %s\n", libname, &strtab[d.d_val]);
			load_shared_lib(chain, &strtab[d.d_val]);
			if(!strncmp(&strtab[d.d_val], "libc.", 5))
				libc_needed = false;
			printf("got it\n");
			break;
		}
	}
	if(libc_needed) {
		printf("getting libc for %s\n", libname);
		load_shared_lib(chain, "libc.so");
		load_shared_lib(chain, "_System_");
	}
	printf("done depend for %s\n", libname);
}


void
Elf32::link_info::initialize()
{
	if(initializing || initialized) 
		THROW(Error("reinitializing library"));
	initializing = true;
	printf("initializing %s\n", libname);
	printf("doing reloc: %lx, %lx\n", relsz, relent);
	for(uint i=0; i<relsz; i += relent) {
		Rel &r = *(Rel*)(rel+i);
		// elf32->dump(r, (Sym*)symtab, strtab);
		reloc(r, loadaddr, bind_now);
	}
	printf("doing reloc2: %lx, %lx\n", relasz, relaent);
	for(uint i=0; i<relasz; i += relaent) {
		Rela &r = *(Rela*)(rel+i);
		// elf32->dump(r, (Sym*)symtab, strtab);
		reloc((Rel)r, r.r_addend, bind_now);
	}
	printf("link table: %p\n", this);
	pltgot[1] = (Addr)this;
	pltgot[2] = (Addr)resolve;
	if(pltrel == DT_REL) 
		for(uint i=0; i<pltrelsz; i += relent) {
			Rel &r = *(Rel*)(jmprel+i);
			// elf32->dump(r, (Sym*)symtab, strtab);
			reloc(r, loadaddr, bind_now);
		}
	else if(pltrel == DT_RELA)  
		for(uint i=0; i<pltrelsz; i += relaent) {
			Rela &r = *(Rela*)(jmprel+i);
			// elf32->dump(r, (Sym*)symtab, strtab);
			reloc((Rel)r, r.r_addend, bind_now);
		}
	printf("calling init: %p\n", init);
	if(init)
		init();
	// XXX/??? -- should we look this up? (atexit) if we're a 
	//  dynamically-loaded dynamic linker, we should,
	//  otherwise, we should already know where it is... (feb)
	if(fini)
		atexit(fini);
	initializing = false;
	initialized = true;
	printf("done %s\n", libname);
}


void
Elf32::load_shared_lib(dyn_chain *chain, const char *name)
{
	link_info *link = chain->link;
	dyn_chain **dynp = &chain->next_level;

	printf("loading %s\n", name);
	while(*dynp) 
		dynp = &(*dynp)->next;
	while(link->prev)
		link = link->prev;
	// link->chain->dump();
	for(link_info *linkp = link; linkp; linkp = linkp->next) {
		printf("checking %s\n", linkp->libname);
		if(linkp->libname && !strcmp(linkp->libname, name)) {
			printf("already loaded at %p\n", linkp);
			*dynp = new dyn_chain(linkp);
			printf("chain:\n");
			// chain->dump();
			return;
		}
	}

	// XXX - should check the cache (/etc/ld.so.cache?) first
	const vfile *newfile;
	vstream *newstream;
	if(!(newfile = thread->get_cwd()->get_file(name, ".:/usr:/lib")))
		THROW(Error("Cannot find library"));
	if(!(newstream = newfile->open())) 
		THROW(Error("Cannot open library"));
	delete newfile;
	Elf32 elf(newstream);
	elf.load();
	printf("loaded, checking name\n");
	for(link_info *linkp = link; linkp; linkp = linkp->next) {
		printf("checking %s\n", linkp->libname);
		if(linkp->libname
		   && !strcmp(linkp->libname, elf.link->libname)) {
			printf("already loaded at %p\n", linkp);
			*dynp = new dyn_chain(linkp);
			printf("chain:\n");
			// chain->dump();
			return;
		}
	}
	printf("got new one, old chain:\n");
	// chain->dump();
	printf("dump done\n");
	*dynp = elf.link->chain;
	printf("a\n");
	elf.link->chain_top = chain->link->chain_top;
	printf("b\n");
	link = chain->link;
	printf("c\n");
	printf("link=%p\n", link);
	while(link->next)
		link = link->next;
	printf("d\n");
	printf("link=%p\n", link);
	link->next = elf.link;
	printf("link=%p\n", link);
	printf("e\n");
	elf.link->prev = link;
	printf("f\n");
	elf.link->next = 0;
	printf("chain top:\n");
	// link->chain_top->dump();
	printf("g\n");
	elf.link->load_dependencies();
}


void
Elf32::attach(link_info *newlink)
{
	newlink->next = link;
	newlink->prev = 0;
	if(link)
		link->prev = newlink;
	link = newlink;
}



// returns address adjustment
Elf32::Addr Elf32::load() 
{
	Addr min = ADDR_MAX;
	Addr max = ADDR_MIN;
	Word size, adjust;
	void *region;

	for(int i=0; i<ehdr.e_phnum; i++) {
		Phdr phdr;
		read(phdr, i);
		if(phdr.p_vaddr < min)
			min = phdr.p_vaddr;
		if(phdr.p_vaddr + phdr.p_memsz > max)
			max = phdr.p_vaddr + phdr.p_memsz;
	}
	size = max-min;
	printf("va: %lx, sz:%lx\n", min, size);
	if(ehdr.e_type == ET_EXEC) 
		printf("WARNING: can't load at specified addresses\n");
	region = (void*)(new char[size]);
	adjust = (int)region - min;
	printf("r:%p adj:%lx\n", region, adjust);
	for(int i=0; i<ehdr.e_phnum; i++) {
		Phdr phdr;
		read(phdr, i);
		if(phdr.p_memsz && phdr.p_filesz) {
			printf("reading %lx at %lx\n",
				phdr.p_memsz, phdr.p_vaddr+adjust);
			stream->read_at(phdr.p_offset,
				(void*)(phdr.p_vaddr+adjust),
				phdr.p_memsz);
		}
		switch(phdr.p_type) {
		case PT_LOAD:
		case PT_PHDR:
			break;
		case PT_DYNAMIC:
			/*
			 if(!first_link) {
				first_link = new link_info(this, 0, _DYNAMIC);
				first_link->dump();
			}
			{
				link_info &link = *(new link_info(this, adjust,
					(Dyn*)(phdr.p_vaddr+adjust)));
				link.prev = first_link;
				link.next = first_link->next;
				first_link->next = &link;
				link.dump();
				link.initialize();
			}
			*/
			{
				link_info *newlink = new link_info(adjust,
					(Dyn*)(phdr.p_vaddr+adjust));
				newlink->chain = new dyn_chain(newlink);
				newlink->chain_top = newlink->chain;
				// newlink->dump();
				attach(newlink);
			}
			break;
		case PT_INTERP:
			printf("interpreter: %s\n",
				(char*)(phdr.p_vaddr+adjust));
			printf("WARNING: Interpreter not supported\n");
			// THROW(Error("Interpreter not supported"));
			break;
		default:
			break;
		}
	}
	printf("loaded\n");
	return adjust;
}


void
Elf32::exec()
{
	Addr adjust;

	adjust = load();
	load_dependencies();
	initialize();
	// link->chain->dump();
	printf("running at %lx/%lx\n", ehdr.e_entry, ehdr.e_entry+adjust);
	((void (*)())(ehdr.e_entry+adjust))();
	printf("done\n");
}

