/*  $Id: File.cc,v 1.5 2000/01/12 03:44:44 shaggy Exp $
 *      relative path: lib/bff/elf32
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/members/shaggy
 */

#pragma Make Includes_Banned += -nostdinc -nostdinc++
#pragma Make Flags_Banned += -static

#pragma Implementation
#define IMPLEMENTATION_File
 
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/param.h>

#include "Error.h"
#include "Perror.h"
#include "File.h"


extern "C" int printf(const char *fmt, ...);

File::File(const char *fname, const char *path)
{
	if(!fname)
		THROW(Error("null filename"));
	if(fname[0] == '/')
		path = "/";
	else if(!path)
		path = ".";
	while(*path) {
		char buf[PATH_MAX];
		char *cp;
		while(*path && (*path <= ' ' || *path == ':' || *path == ';'))
			path++;
		if(!*path)
			break;
		for(cp = buf; *path > ' ' && *path != ':'
				&& *path != ';'; cp++, path++) 
			*cp = *path;
		--cp;
		if(*cp++ != '/')
			*cp++ = '/';
		strcpy(cp, fname);
		if((_fd = open(buf, O_RDONLY)) >= 0) {
			printf("opened %s\n", buf);
			return;
		}
	}
	THROW(Perror(fname));
}

 
File::~File()
{
	if(close(_fd) < 0)
		THROW(Perror(_fd));
}

 
ssize_t File::read(DStr buf) const
{
	int n = ::read(_fd, buf.ptr, buf.len);
	if(n < 0)
		THROW(Perror(buf.len));
	return n;
}


ssize_t File::write(CStr buf) const
{
	int n = ::write(_fd, buf.ptr, buf.len);
	if(n < 0)
		THROW(Perror(buf.len));
	return n; 
}


off_t File::seek(off_t p) const
{
	off_t n = ::lseek(_fd, p, SEEK_SET);
	if(n < 0)
		THROW(Perror(p));
	return n; 
}


off_t File::tell() const
{
	off_t n = ::lseek(_fd, 0, SEEK_CUR);
	if(n < 0)
		THROW(Perror(_fd));
	return n;
}


ssize_t File::read_at(off_t p, DStr buf) const
{
	seek(p);
	return read(buf);
}


ssize_t File::write_at(off_t p, CStr buf) const
{
	seek(p);
	return write(buf); 
};


