/*  $Id: vdir.cc,v 1.8 2000/01/12 03:44:44 shaggy Exp $
 *      relative path: lib/c++
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/members/shaggy
 */

#include <libc_port.h>
#include <stdx1.h>
#include <vobject>
#include <ParsePath>
#include <vfile>
#include <vthread>
#include <vdirstream>

#include "vdir.hh"

extern "C" int printf(const char *fmt, ...);

const vobject*
vdir::_get_object(const char *name) const
{
const vobject *vobj = 0;
vdirstream *dirstream = opendir();
	if(!dirstream)
		return 0;
	while(dirstream->read())  {
		// printf("matching (%s)\n", dirstream->get_name());
		if(dirstream->match(name)) {
			vobj = dirstream->get_object();
			break;
		}
	}
	delete dirstream;
	// if(vobj)
	//	printf("returning object (%s)\n", vobj->get_path());
	// else
	//	printf("no object\n");
	return vobj;
}


// XXX -- currently, search paths are not supported.
// should we support search paths?
// or should we use search virtual directories? (with links)
// seach path lists (with ':' separation) can be easily
// converted into search directories, which involves more
// initial overhead, but makes further searching faster, locally at least.
// However, over remote file systems, giving a full path might be faster...

 
const vobject*
vdir::get_object(const char *fname, const char *) const
{
ParsePath path(fname);
const vdir *dir;
char name[NAME_MAX];
	// printf("getting object %s from (%s)\n", fname, get_path());
	if(path.is_absolute())
		dir = thread->get_root();
	else
		dir = this;
	// printf("starting from (%s)\n", dir->get_path());
	while(path.get_next_dir(name)) 
		if(!(dir = dir->_get_dir(name))) 
			return 0;
	if(path.has_base())
		return dir->_get_object(path.get_base());
	else
		return dir;
}

const vfile*
vdir::_get_file(const char *name) const
{
const vobject *vobj = _get_object(name);
	if(vobj && vobj->is_a(class_vfile)) return (const vfile*)vobj;
	else return 0; 
}

const vfile*
vdir::get_file(const char *fname, const char *path) const
{
const vobject *vobj = get_object(fname, path);
	if(vobj && vobj->is_a(class_vfile)) return (const vfile*)vobj;
	else return 0; 
}

const vdir*
vdir::_get_dir(const char *name) const
{
const vobject *vobj = _get_object(name);
	if(vobj && vobj->is_a(class_vdir)) return (const vdir*)vobj;
	else return 0;
}

const vdir*
vdir::get_dir(const char *fname, const char *path) const
{
const vobject *vobj = get_object(fname, path);
	if(vobj && vobj->is_a(class_vdir)) return (const vdir*)vobj;
	else return 0; 
}

vstream*
vdir::_open(const char *name, int flags) const
{
const vfile *file = _get_file(name);
	if(file) return file->open(flags);
	else return 0;
}

vstream*
vdir::open(const char *fname, int flags) const
{
const vfile *file = get_file(fname, 0);
	if(file) return file->open(flags);
	else return 0;
}

vdirstream*
vdir::_opendir(const char *name) const
{
const vdir *dir = _get_dir(name);
	if(dir) return dir->opendir();
	else return 0;
}

vdirstream*
vdir::opendir(const char *fname) const
{
const vdir *dir = get_dir(fname, 0);
	if(dir) return dir->opendir();
	else return 0;
}

bool
vdir::_stat(const char *name, vstat *vs) const
{
const vobject *vobj = _get_object(name);
	if(vobj) return vobj->stat(vs);
	else return false;
}

bool
vdir::stat(const char *fname, vstat *vs) const
{
const vobject *vobj = get_object(fname, 0);
	if(vobj) return vobj->stat(vs);
	else return false;
}


bool
vdir::_mkdir(const char *) const
{
	return false;
}

bool
vdir::mkdir(const char *dname) const
{
ParsePath path(dname);
const vdir *dir;
char name[NAME_MAX];
	if(!path.has_base())
		return false;
	if(path.is_absolute())
		dir = thread->get_root();
	else
		dir = this;
	while(path.get_next_dir(name)) {
		if(!(dir = dir->_get_dir(name))) 
			return false;
	}
	return _mkdir(path.get_base());
}

