/*  $Id: boot.h,v 1.20 1999/07/16 20:18:51 shaggy Exp $
 *      Boot Info for SOLO/Shagboot
 *      relative path: ./sys/boot/m-pc
 */ 

#include "chunk.h"


#define SECTOR_SIZE		0x200	// size of a disk sector 
#define SECTOR_2SIZE		9	// shift count for mul/div

#define PARTID_UNUSED	0x00
#define PARTID_DOS12	0x01
#define PARTID_EXTENDED	0x05

#define	BOOT_MAGIC		0xaa55
#define	BOOTMERGE_MAGIC		0xaa66


#define MAP_SIZE		0x1800		// 6K max


#ifdef ASM

#include "boot.ofs"

#if BootBlock__sizeof != 0x200
#error sizeof(BootBlock) != 0x200 (see boot.ofs)
#error Perhaps the compiler is failing to pack structures correctly?
#endif


// important addresses:

#define	PTAB_ADR	0x07be		// where to copy partition table
#define BOOT_ADR	0x7c00		// where boot sector is located
// although it's supposed to be safe to start using memory at 0x0c00,
// (and previous versions of SOLO resided entirely in the first 32K)
// skipping out of the first 24K altogether allows more boot schemes
// to coexist, since some of them want to use special areas down here.
// If you really want to reclaim this space, your OS can always grab it
// back, but by leaving it alone, it's there for any other special purposes
// you might have for it in the future, to allow extra information to be
// passed to the kernel, or to survive resets, or for compatibility with
// other bootloaders that expect to put particular bits of information here.

#define VIDMODE_ADR	0x1000		// where to keep VESA-style mode info
#define VIDMODE_SIZE		0x100		// 256 bytes
// 0x1100-0x11ff is reserved for future purposes
// 0x1200-0x1dff is reserved for compatibility with other boot loaders
#define VIDINFO_ADR	0x1e00		// where to keep VESA-style card info
#define VIDINFO_SIZE		0x200		// 512 bytes
// 0x2000-0x3fff is reserved for compatibility with other boot loaders
#define BUF_ADR		0x4000		// BPMI call buffer area
#define BUF_SIZE		0x4000		// 16K max (incl. bootsec)
#define LOAD_ADR	0x8000		// where to load SOLO 
#define LOAD_SIZE		0x4400		// 17K max (incl. cmdline/env)
/* XXX - note that 1K is unused here, due to a bug in the environment var
 * code... once that is fixed, the load size can be increased to 18K
 */
#define FDT_ADR		0xc800 		// file descriptor table
#define FDT_SIZE		0x0180		// 384 max (12 FD's)
#define TSS_ADR		0xc980		// where to keep TSS info
#define TSS_SIZE		0x0080		// 128 max
#define TMP_ADR		0xca00		// temporary buffer area
#define TMP_SIZE		0x0200		// 512 max (one sector)
#define ARGV_ADR	0xcc00		// argv[] array
#define ARGV_SIZE		0x0100		// 256 max (64 args)
#define PARAM_ADR	0xcd00		// where to put BootParam block
#define PARAM_SIZE		0x0100		// 256 max
#define LINE_ADR	0xce00		// input line buffer
#define	LINE_SIZE		0x0100		// 256 max (252 if debug)
#define STACK_ADR	0xcf00		// stack: 0xcf00 - 0xcfff
#define STACK_SIZE		0x0100		// 256 max
#define MAP_ADR		0xd000		// where to load map file 
//#define MAP_SIZE		0x1800		defined up above (6K max)
#define PROGRESS	0xe7fe	// used for keeping track of
				// progress in case of crash (debugging)
#define MSG_ADR		0xe800		// area for kernel messages
#define MSG_SIZE		0x0800		// 2K max
#define INIT_ADR	0xf000		// start address for executables
#define EXT_ADR		0x00100000	// start of extended memory

#define CMDLINE_SIZE	0x200		// 512 - max size of command line,
					// after expansion 
#define STACK_PTR	STACK_ADR+STACK_SIZE

#define SCRIPT_ADR	0x20000		// where to cache scripts
#define SOLOEXT_ADR	0x30000		// where to cache extensions


#define BOOTCODE_ADR		0x7c40		// BOOT_ADR + BOOTBLOCK_CODE


// Note:
//	6K for the map file area means that there can
//	be up to slightly less than 1536 (depending on how
//	many file names are there) chunks.
//	each chunk can map anywhere from 1 to 64 sectors,
//	so the mapfile can map from 768K to 48MB of startup files,
//	which should be sufficient for most cases.
//	To save mapfile space, (if you're running out)
//	make sure the disk being mapped isn't badly fragmented.
//	If necessary, the boundaries can be moved around, but
//	remember that any changes will cause incompatibilities
//	between ShagBoot and SOLO unless both are recompiled
//	and reinstalled.

//	Note that, in the worse case, with 512-byte blocks,
//	and complete fragmentation, loading the full 6K
//	mapfile will require 12 chunks.  Currently only 8
//	chunks are reserved in the boot sector, so this
//	could potentially be a problem, although it is
//	highly unlikely.  If you run into this problem,
//	please contact me.

// Why does all this stuff have to fit in the lower 64K of memory?
// you may ask...
// Well, notice that it all stays resident while the kernel is running,
// to be available for BIOS calls, static file system access,
// and for debugging in case the kernel crashes.  Therefore, it
// is important to keep it as small as possible, and by keeping
// it within the first 60K, the next 4K is available for loading
// the headers of executables, while keeping everything in the
// first 64K segment, so that no messing around with segment
// registers has to be done, thus keeping it even smaller and simpler.

#else

#include <stdx1.h>

#ifdef __notemplates
	typedef	uint16		packed_uint16;
	typedef uint32		packed_uint32;
#else
#	include <packed>
	typedef packed<uint16>	packed_uint16;
	typedef packed<uint32>	packed_uint32;
#endif

// XXX - see the note at the bottom
//	all of this code had to be drastically rewritten
//	for GCC 2.7.2 since __attribute__((packed)) doesn't
//	seem to work anymore.

// packed_uint16 and packed_uint8 were created to overcome some
// of the problems, and bitfields had to be replaced with inline methods
// which means that code had to be modified to use the new inline methods
// instead of going directly to the bitfields.
// I suppose that's even more reason to do everything in the future
// through inline methods rather than accessing structures directly
// from higher level code.


struct DiskParam {
	packed_uint16	cylinders;
	uint8	heads;
	packed_uint16	startcyl;
	packed_uint16	writeprecomp;
	uint8 	maxecc;
	uint8 	control;
	uint8 	timeout;
	uint8	ftimeout;
	uint8 	ctimeout;
	packed_uint16 	landing;
	uint8 	sectors;
	uint8 	unused;
};

struct Partition {
	uint8 	boot;
	uint8	starthead;	// top 2 bits are bits 10-11 of startcylinder
	uint8	startsector;	// top 2 bits are bits 8-9 of startcylinder
	uint8	startcylinder;
	uint8 	id;
	uint8	endhead;	// top 2 bits are bits 10-11 of endcylinder
	uint8	endsector;	// top 2 bits are bits 8-9 of endcylinder
	uint8	endcylinder;
	packed_uint32	start;	// linear sector address
	packed_uint32	size;	// linear sector size

// XXX - well, I started to implement access methods for all of 
//	 what used to be bitfields, and then, upon searching for
//	 where they were used in the C++ code, discovered 
//	 that they aren't!
//	 So, when/if they do get used, the rest of these
//	 will have to be implemented, but for now, I have
//	 more important places to focus my time and effort...

//	Actually, this stuff might turn out not to be necessary,
//	since it seems that you can declare bitfields with
//	smaller alignments with (i.e.)  uchar bit:3;
//	(unfortunately enum types always align to int,
//	except with -fshort-enum)
//	However, access methods to take care of
//	the "scattered bits for backwards compatibility" problem
//	might still be a good idea, if this is ever needed.

/*
	uint start_sector() { return get_bits(startsector, 0, 6); }
	void start_sector(uint n) { set_bits(startsector, 0, 6, n); }
	uint start_cyl() { return startcylinder
				| (get_bits(startsector, 6, 2) << 6)
				| (get_bits(starthead, 6, 2) << 8); }
*/
};


typedef struct BootBlock BootBlock;
struct BootBlock {
	uint8		jmp[3];		// skip over header stuff
	char		sysname[8];	// system name for DOS
	uint8		bpb[0x2c];	// Bios Parameter Block
	uint8		resv[5];	// 16-byte alignment padding
	packed_uint32	bootsec;	// dev/head/cyl/sec for boot sector
	uint8		code[0x14a]; 	// up to 330 bytes of code
	packed_uint16	bootoffset;	// offset of boot program in map
	Chunk		chunk[8];	// up to 8 chunks 
	uint8		mapsize;	// # of map chunks
	uint8 		mapdev;		// device ID for map
	DiskParam	diskparam __attribute__ ((packed)) ;
	Partition	partition[4] __attribute__ ((packed)) ;
	packed_uint16	magic;
};


#endif

