/*  $Id: lsboot.cc,v 1.12 2000/01/12 03:44:47 shaggy Exp $
 *      List all bootable devices and files
 *      relative path: sys/boot/m-pc
 *  
 *  Copyright (c) 1996-1999 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/members/shaggy
 */

#define lsboot_VER_MAJOR	0
#define lsboot_VER_MINOR	2
#define lsboot_VER_PATCH	0
#define lsboot_VER_STR	"0.2"
#define lsboot_VER_DATE	"January 20, 1997"
#define lsboot_VER_NDATE	"01/20/97"


#include <libc.h>
#include <stdx1.h>
#include <defines>
#include <boot/bootparam.h>
#include <boot/boot.h>
#include <boot/device.h>
#include <boot/mapfile.h>


static csz rcsid[] __attribute__((unused))
	= "$Id: lsboot.cc,v 1.12 2000/01/12 03:44:47 shaggy Exp $"
	  "$Release: " lsboot_VER_STR " $";


extern BootParam *bootparam;

int
lsboot(int /*argc*/, csz **/*argv*/)
{
	printf("Bootable devices, partitions, and files:\n");
	printf("[name]           [text found in boot sector]\n");

	const MapFile* mapfile;
	int i = 0;

	while((mapfile = bootparam->readdir(i++))) {
		if(mapfile->device & DEVICE_MAPPED) {
			if(mapfile->fsize != SECTOR_SIZE)
				continue;
		} else {
			if(mapfile->device != DEV_DISK)
				continue;
		}
		BootParam::fd_t fd = bootparam->open(mapfile->name);
		BootBlock sector;
		size_t len = bootparam->read(fd, &sector, sizeof(sector));
		bootparam->close(fd);
		if(len != sizeof(sector))
			continue;
		if(sector.magic != BOOT_MAGIC
				&& sector.magic != BOOTMERGE_MAGIC) 
			continue;
		sz buf[63];
		sz *cp = buf;
		if(!sector.jmp[0])	
			cp += sprintf(cp, "<invalid boot code> ");
		csz *sp = (csz*)&sector.sysname;
		int count = 0;

		// extract as much readable text as possible from
		// the boot sector, to identify it:

		while(cp-buf < (int)sizeof(buf)-2
				&& sp-(csz*)&sector < (int)sizeof(sector)) {
			if(*sp >= 0x20 && *sp <= 0x7a)
				count++;
			else {
				if(count >= 3 && cp[-1] != ' ')
					*cp++ = ' ';
				count = 0;
			}
			if(count == 3) {
				if(sp[-2]=='P' && sp[-1]=='S'
						&& sp[0] == 'R')
					count = 0;
				else {
					*cp++ = sp[-2];
					*cp++ = sp[-1];
				}
			}
			if(count >= 3 && (*sp != ' ' || cp[-1] != ' '))
				*cp++ = *sp;
			sp++;
		}

		*cp = '\0';
		printf("%-16s %s\n", mapfile->name, buf);
	}
	return 0;
}

